# -*- coding: utf-8 -*-
# auth.py
# Copyright (C) 2013 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Twisted http token auth.
"""
import binascii
import time

from hashlib import sha512
from zope.interface import implementer

from twisted.cred import error
from twisted.cred.checkers import ICredentialsChecker
from twisted.cred.credentials import IUsernamePassword
from twisted.cred.credentials import IAnonymous
from twisted.cred.credentials import Anonymous
from twisted.cred.credentials import UsernamePassword
from twisted.cred.portal import IRealm
from twisted.cred.portal import Portal
from twisted.internet import defer
from twisted.logger import Logger
from twisted.web.iweb import ICredentialFactory
from twisted.web.resource import IResource

from leap.soledad.common.couch import couch_server

from ._resource import SoledadResource, SoledadAnonResource
from ._blobs import BlobsResource
from ._config import get_config


log = Logger()


@implementer(IRealm)
class SoledadRealm(object):

    def __init__(self, sync_pool, conf=None):
        assert sync_pool is not None
        if conf is None:
            conf = get_config()
        blobs = conf['blobs']
        blobs_resource = BlobsResource("filesystem",
                                       conf['blobs_path']) if blobs else None
        self.anon_resource = SoledadAnonResource(
            enable_blobs=blobs)
        self.auth_resource = SoledadResource(
            blobs_resource=blobs_resource,
            sync_pool=sync_pool)

    def requestAvatar(self, avatarId, mind, *interfaces):

        # Anonymous access
        if IAnonymous.providedBy(avatarId):
            return (IResource, self.anon_resource,
                    lambda: None)

        # Authenticated access
        else:
            if IResource in interfaces:
                return (IResource, self.auth_resource,
                        lambda: None)
        raise NotImplementedError()


@implementer(ICredentialsChecker)
class TokenChecker(object):

    credentialInterfaces = [IUsernamePassword, IAnonymous]

    TOKENS_DB_PREFIX = "tokens_"
    TOKENS_DB_EXPIRE = 30 * 24 * 3600  # 30 days in seconds
    TOKENS_TYPE_KEY = "type"
    TOKENS_TYPE_DEF = "Token"
    TOKENS_USER_ID_KEY = "user_id"

    def __init__(self):
        self._couch_url = get_config().get('couch_url')

    def _get_server(self):
        return couch_server(self._couch_url)

    def _tokens_dbname(self):
        # the tokens db rotates every 30 days, and the current db name is
        # "tokens_NNN", where NNN is the number of seconds since epoch
        # divide dby the rotate period in seconds. When rotating, old and
        # new tokens db coexist during a certain window of time and valid
        # tokens are replicated from the old db to the new one. See:
        # https://leap.se/code/issues/6785
        dbname = self.TOKENS_DB_PREFIX + \
            str(int(time.time() / self.TOKENS_DB_EXPIRE))
        return dbname

    def _tokens_db(self):
        dbname = self._tokens_dbname()

        # TODO -- leaking abstraction here: this module shouldn't need
        # to known anything about the context manager. hide that in the couch
        # module
        with self._get_server() as server:
            db = server[dbname]
        return db

    def requestAvatarId(self, credentials):
        if IAnonymous.providedBy(credentials):
            return defer.succeed(Anonymous())

        uuid = credentials.username
        token = credentials.password

        # lookup key is a hash of the token to prevent timing attacks.
        # TODO cache the tokens already!

        db = self._tokens_db()
        token = db.get(sha512(token).hexdigest())
        if token is None:
            return defer.fail(error.UnauthorizedLogin())

        # TODO -- use cryptography constant time builtin comparison.
        # we compare uuid hashes to avoid possible timing attacks that
        # might exploit python's builtin comparison operator behaviour,
        # which fails immediatelly when non-matching bytes are found.
        couch_uuid_hash = sha512(token[self.TOKENS_USER_ID_KEY]).digest()
        req_uuid_hash = sha512(uuid).digest()
        if token[self.TOKENS_TYPE_KEY] != self.TOKENS_TYPE_DEF \
                or couch_uuid_hash != req_uuid_hash:
            return defer.fail(error.UnauthorizedLogin())

        return defer.succeed(uuid)


@implementer(ICredentialFactory)
class TokenCredentialFactory(object):

    scheme = 'token'

    def getChallenge(self, request):
        return {}

    def decode(self, response, request):
        try:
            creds = binascii.a2b_base64(response + b'===')
        except binascii.Error:
            raise error.LoginFailed('Invalid credentials')

        creds = creds.split(b':', 1)
        if len(creds) == 2:
            return UsernamePassword(*creds)
        else:
            raise error.LoginFailed('Invalid credentials')


def portalFactory(sync_pool):
    realm = SoledadRealm(sync_pool=sync_pool)
    checker = TokenChecker()
    return Portal(realm, [checker])


credentialFactory = TokenCredentialFactory()



# -*- coding: utf-8 -*-
# interfaces.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


from zope.interface import Interface


class IBlobsBackend(Interface):

    """
    An interface for a BlobsBackend.
    """

    def read_blob(user, blob_id, request):
        """
        Read blob with a given blob_id, and write it to the passed request.

        :returns: a deferred that fires upon finishing.
        """

    def write_blob(user, blob_id, request):
        """
        Write blob to the storage, reading it from the passed request.

        :returns: a deferred that fires upon finishing.
        """

    def delete_blob(user, blob_id):
        """
        Delete the given blob_id.
        """

    def get_blob_size(user, blob_id):
        """
        Get the size of the given blob id.
        """

    def list_blobs(user, request):
        """
        Returns a json-encoded list of ids from user's blob.

        :returns: a deferred that fires upon finishing.
        """

    def get_total_storage(user):
        """
        Get the size used by a given user as the sum of all the blobs stored
        unders its namespace.
        """

    def add_tag_header(user, blob_id, request):
        """
        Adds a header 'Tag' to the passed request object, containing the last
        16 bytes of the encoded blob, which according to the spec contains the
        tag.

        :returns: a deferred that fires upon finishing.
        """



# -*- coding: utf-8 -*-
# sync.py
# Copyright (C) 2014 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
Server side synchronization infrastructure.
"""
import time
from six.moves import zip as izip

from leap.soledad.common.l2db import sync
from leap.soledad.common.l2db.remote import http_app
from leap.soledad.server.caching import get_cache_for
from leap.soledad.server.state import ServerSyncState
from leap.soledad.common.document import ServerDocument


MAX_REQUEST_SIZE = float('inf')  # It's a stream.
MAX_ENTRY_SIZE = 200  # in Mb
ENTRY_CACHE_SIZE = 8192 * 1024


class SyncExchange(sync.SyncExchange):

    def __init__(self, db, source_replica_uid, last_known_generation, sync_id):
        """
        :param db: The target syncing database.
        :type db: SoledadBackend
        :param source_replica_uid: The uid of the source syncing replica.
        :type source_replica_uid: str
        :param last_known_generation: The last target replica generation the
                                      source replica knows about.
        :type last_known_generation: int
        :param sync_id: The id of the current sync session.
        :type sync_id: str
        """
        self._db = db
        self.source_replica_uid = source_replica_uid
        self.source_last_known_generation = last_known_generation
        self.sync_id = sync_id
        self.new_gen = None
        self.new_trans_id = None
        self._trace_hook = None
        # recover sync state
        self._sync_state = ServerSyncState(self.source_replica_uid, sync_id)

    def find_changes_to_return(self):
        """
        Find changes to return.

        Find changes since last_known_generation in db generation
        order using whats_changed. It excludes documents ids that have
        already been considered (superseded by the sender, etc).

        :return: the generation of this database, which the caller can
                 consider themselves to be synchronized after processing
                 allreturned documents, and the amount of documents to be sent
                 to the source syncing replica.
        :rtype: int
        """
        # check if changes to return have already been calculated
        new_gen, new_trans_id, number_of_changes = self._sync_state.sync_info()
        if number_of_changes is None:
            self._trace('before whats_changed')
            new_gen, new_trans_id, changes = self._db.whats_changed(
                self.source_last_known_generation)
            self._trace('after whats_changed')
            seen_ids = self._sync_state.seen_ids()
            # changed docs that weren't superseded by or converged with
            self.changes_to_return = [
                (doc_id, gen, trans_id) for (doc_id, gen, trans_id) in changes
                # there was a subsequent update
                if doc_id not in seen_ids or seen_ids.get(doc_id) < gen]
            self._sync_state.put_changes_to_return(
                new_gen, new_trans_id, self.changes_to_return)
            number_of_changes = len(self.changes_to_return)
        self.new_gen = new_gen
        self.new_trans_id = new_trans_id
        return self.new_gen, number_of_changes

    def return_docs(self, return_doc_cb):
        """Return the changed documents and their last change generation
        repeatedly invoking the callback return_doc_cb.

        The final step of a sync exchange.

        :param: return_doc_cb(doc, gen, trans_id): is a callback
                used to return the documents with their last change generation
                to the target replica.
        :return: None
        """
        changes_to_return = self.changes_to_return
        # return docs, including conflicts.
        # content as a file-object (will be read when writing)
        changed_doc_ids = [doc_id for doc_id, _, _ in changes_to_return]
        docs = self._db.get_docs(
            changed_doc_ids, check_for_conflicts=False,
            include_deleted=True, read_content=False)

        docs_by_gen = izip(
            docs, (gen for _, gen, _ in changes_to_return),
            (trans_id for _, _, trans_id in changes_to_return))
        for doc, gen, trans_id in docs_by_gen:
            return_doc_cb(doc, gen, trans_id)

    def batched_insert_from_source(self, entries, sync_id):
        if not entries:
            return
        self._db.batch_start()
        for entry in entries:
            doc, gen, trans_id, number_of_docs, doc_idx = entry
            self.insert_doc_from_source(doc, gen, trans_id, number_of_docs,
                                        doc_idx, sync_id)
        self._db.batch_end()

    def insert_doc_from_source(
            self, doc, source_gen, trans_id,
            number_of_docs=None, doc_idx=None, sync_id=None):
        """Try to insert synced document from source.

        Conflicting documents are not inserted but will be sent over
        to the sync source.

        It keeps track of progress by storing the document source
        generation as well.

        The 1st step of a sync exchange is to call this repeatedly to
        try insert all incoming documents from the source.

        :param doc: A Document object.
        :type doc: Document
        :param source_gen: The source generation of doc.
        :type source_gen: int
        :param trans_id: The transaction id of that document change.
        :type trans_id: str
        :param number_of_docs: The total amount of documents sent on this sync
                               session.
        :type number_of_docs: int
        :param doc_idx: The index of the current document.
        :type doc_idx: int
        :param sync_id: The id of the current sync session.
        :type sync_id: str
        """
        state, at_gen = self._db._put_doc_if_newer(
            doc, save_conflict=False, replica_uid=self.source_replica_uid,
            replica_gen=source_gen, replica_trans_id=trans_id,
            number_of_docs=number_of_docs, doc_idx=doc_idx, sync_id=sync_id)
        if state == 'inserted':
            self._sync_state.put_seen_id(doc.doc_id, at_gen)
        elif state == 'converged':
            # magical convergence
            self._sync_state.put_seen_id(doc.doc_id, at_gen)
        elif state == 'superseded':
            # we have something newer that we will return
            pass
        else:
            # conflict that we will returne
            assert state == 'conflicted'


class SyncResource(http_app.SyncResource):

    max_request_size = MAX_REQUEST_SIZE * 1024 * 1024
    max_entry_size = MAX_ENTRY_SIZE * 1024 * 1024

    sync_exchange_class = SyncExchange

    @http_app.http_method(
        last_known_generation=int, last_known_trans_id=http_app.none_or_str,
        sync_id=http_app.none_or_str, content_as_args=True)
    def post_args(self, last_known_generation, last_known_trans_id=None,
                  sync_id=None, ensure=False):
        """
        Handle the initial arguments for the sync POST request from client.

        :param last_known_generation: The last server replica generation the
                                      client knows about.
        :type last_known_generation: int
        :param last_known_trans_id: The last server replica transaction_id the
                                    client knows about.
        :type last_known_trans_id: str
        :param sync_id: The id of the current sync session.
        :type sync_id: str
        :param ensure: Whether the server replica should be created if it does
                       not already exist.
        :type ensure: bool
        """
        # create or open the database
        cache = get_cache_for('db-' + sync_id + self.dbname, expire=120)
        if ensure:
            db, self.replica_uid = self.state.ensure_database(self.dbname)
        else:
            db = self.state.open_database(self.dbname)
        db.init_caching(cache)
        # validate the information the client has about server replica
        db.validate_gen_and_trans_id(
            last_known_generation, last_known_trans_id)
        # get a sync exchange object
        self.sync_exch = self.sync_exchange_class(
            db, self.source_replica_uid, last_known_generation, sync_id)
        self._sync_id = sync_id
        self._staging = []
        self._staging_size = 0

    @http_app.http_method(content_as_args=True)
    def post_put(
            self, id, rev, content, gen,
            trans_id, number_of_docs, doc_idx):
        """
        Put one incoming document into the server replica.

        :param id: The id of the incoming document.
        :type id: str
        :param rev: The revision of the incoming document.
        :type rev: str
        :param content: The content of the incoming document.
        :type content: dict
        :param gen: The source replica generation corresponding to the
                    revision of the incoming document.
        :type gen: int
        :param trans_id: The source replica transaction id corresponding to
                         the revision of the incoming document.
        :type trans_id: str
        :param number_of_docs: The total amount of documents sent on this sync
                               session.
        :type number_of_docs: int
        :param doc_idx: The index of the current document.
        :type doc_idx: int
        """
        doc = ServerDocument(id, rev, json=content)
        self._staging_size += len(content or '')
        self._staging.append((doc, gen, trans_id, number_of_docs, doc_idx))
        if self._staging_size > ENTRY_CACHE_SIZE or doc_idx == number_of_docs:
            self.sync_exch.batched_insert_from_source(self._staging,
                                                      self._sync_id)
            self._staging = []
            self._staging_size = 0

    def post_get(self):
        """
        Return syncing documents to the client.
        """
        def send_doc(doc, gen, trans_id):
            entry = dict(id=doc.doc_id, rev=doc.rev,
                         gen=gen, trans_id=trans_id)
            self.responder.stream_entry(entry)
            content_reader = doc.get_json()
            if content_reader:
                content = content_reader.read()
                self.responder.stream_entry(content)
                content_reader.close()
                # throttle at 5mb/s
                # FIXME: twistd cant control througput
                # we need to either use gunicorn or go async
                time.sleep(len(content) / (5.0 * 1024 * 1024))
            else:
                self.responder.stream_entry('')

        new_gen, number_of_changes = \
            self.sync_exch.find_changes_to_return()
        self.responder.content_type = 'application/x-u1db-sync-response'
        self.responder.start_response(200)
        self.responder.start_stream(),
        header = {
            "new_generation": new_gen,
            "new_transaction_id": self.sync_exch.new_trans_id,
            "number_of_changes": number_of_changes,
        }
        if self.replica_uid is not None:
            header['replica_uid'] = self.replica_uid
        self.responder.stream_entry(header)
        self.sync_exch.return_docs(send_doc)
        self.responder.end_stream()
        self.responder.finish_response()

    def post_end(self):
        """
        Return the current generation and transaction_id after inserting one
        incoming document.
        """
        self.responder.content_type = 'application/x-soledad-sync-response'
        self.responder.start_response(200)
        self.responder.start_stream(),
        new_gen, new_trans_id = self.sync_exch._db._get_generation_info()
        header = {
            "new_generation": new_gen,
            "new_transaction_id": new_trans_id,
        }
        if self.replica_uid is not None:
            header['replica_uid'] = self.replica_uid
        self.responder.stream_entry(header)
        self.responder.end_stream()
        self.responder.finish_response()



# -*- coding: utf-8 -*-
# _blobs.py
# Copyright (C) 2017 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
Blobs Server implementation.

This is a very simplistic implementation for the time being.
Clients should be able to opt-in util the feature is complete.

A more performant BlobsBackend can (and should) be implemented for production
environments.
"""
import os
import base64
import json
import re

from twisted.logger import Logger
from twisted.web import static
from twisted.web import resource
from twisted.web.client import FileBodyProducer
from twisted.web.server import NOT_DONE_YET
from twisted.internet import utils, defer

from zope.interface import implementer

from leap.common.files import mkdir_p
from leap.soledad.server import interfaces


__all__ = ['BlobsResource']


logger = Logger()

# Used for sanitizers, we accept only letters, numbers, '-' and '_'
VALID_STRINGS = re.compile('^[a-zA-Z0-9_-]+$')


# for the future:
# [ ] isolate user avatar in a safer way
# [ ] catch timeout in the server (and delete incomplete upload)
# [ ] chunking (should we do it on the client or on the server?)


@implementer(interfaces.IBlobsBackend)
class FilesystemBlobsBackend(object):

    def __init__(self, blobs_path='/tmp/blobs/', quota=200 * 1024):
        self.quota = quota
        if not os.path.isdir(blobs_path):
            os.makedirs(blobs_path)
        self.path = blobs_path

    def read_blob(self, user, blob_id, request):
        logger.info('reading blob: %s - %s' % (user, blob_id))
        path = self._get_path(user, blob_id)
        logger.debug('blob path: %s' % path)
        _file = static.File(path, defaultType='application/octet-stream')
        return _file.render_GET(request)

    @defer.inlineCallbacks
    def write_blob(self, user, blob_id, request):
        path = self._get_path(user, blob_id)
        try:
            mkdir_p(os.path.split(path)[0])
        except OSError:
            pass
        if os.path.isfile(path):
            # 409 - Conflict
            request.setResponseCode(409)
            request.write("Blob already exists: %s" % blob_id)
            defer.returnValue(None)
        used = yield self.get_total_storage(user)
        if used > self.quota:
            logger.error("Error 507: Quota exceeded for user: %s" % user)
            request.setResponseCode(507)
            request.write('Quota Exceeded!')
            defer.returnValue(None)
        logger.info('writing blob: %s - %s' % (user, blob_id))
        fbp = FileBodyProducer(request.content)
        yield fbp.startProducing(open(path, 'wb'))

    def delete_blob(self, user, blob_id):
        blob_path = self._get_path(user, blob_id)
        os.unlink(blob_path)

    def get_blob_size(user, blob_id):
        raise NotImplementedError

    def list_blobs(self, user, request):
        blob_ids = []
        base_path = self._get_path(user)
        for _, _, filenames in os.walk(base_path):
            blob_ids += filenames
        return json.dumps(blob_ids)

    def get_total_storage(self, user):
        return self._get_disk_usage(self._get_path(user))

    def add_tag_header(self, user, blob_id, request):
        with open(self._get_path(user, blob_id)) as doc_file:
            doc_file.seek(-16, 2)
            tag = base64.urlsafe_b64encode(doc_file.read())
            request.responseHeaders.setRawHeaders('Tag', [tag])

    @defer.inlineCallbacks
    def _get_disk_usage(self, start_path):
        if not os.path.isdir(start_path):
            defer.returnValue(0)
        cmd = ['/usr/bin/du', '-s', '-c', start_path]
        output = yield utils.getProcessOutput(cmd[0], cmd[1:])
        size = output.split()[0]
        defer.returnValue(int(size))

    def _validate_path(self, desired_path, user, blob_id):
        if not VALID_STRINGS.match(user):
            raise Exception("Invalid characters on user: %s" % user)
        if blob_id and not VALID_STRINGS.match(blob_id):
            raise Exception("Invalid characters on blob_id: %s" % blob_id)
        desired_path = os.path.realpath(desired_path)  # expand path references
        root = os.path.realpath(self.path)
        if not desired_path.startswith(root + os.sep + user):
            err = "User %s tried accessing a invalid path: %s" % (user,
                                                                  desired_path)
            raise Exception(err)
        return desired_path

    def _get_path(self, user, blob_id=False):
        parts = [user]
        if blob_id:
            parts += [blob_id[0], blob_id[0:3], blob_id[0:6]]
            parts += [blob_id]
        path = os.path.join(self.path, *parts)
        return self._validate_path(path, user, blob_id)


class ImproperlyConfiguredException(Exception):
    pass


class BlobsResource(resource.Resource):

    isLeaf = True

    # Allowed backend classes are defined here
    handlers = {"filesystem": FilesystemBlobsBackend}

    def __init__(self, backend, blobs_path, **backend_kwargs):
        resource.Resource.__init__(self)
        self._blobs_path = blobs_path
        backend_kwargs.update({'blobs_path': blobs_path})
        if backend not in self.handlers:
            raise ImproperlyConfiguredException("No such backend: %s", backend)
        self._handler = self.handlers[backend](**backend_kwargs)
        assert interfaces.IBlobsBackend.providedBy(self._handler)

    # TODO double check credentials, we can have then
    # under request.

    def render_GET(self, request):
        logger.info("http get: %s" % request.path)
        user, blob_id = self._validate(request)
        if not blob_id:
            return self._handler.list_blobs(user, request)
        self._handler.add_tag_header(user, blob_id, request)
        return self._handler.read_blob(user, blob_id, request)

    def render_DELETE(self, request):
        logger.info("http put: %s" % request.path)
        user, blob_id = self._validate(request)
        self._handler.delete_blob(user, blob_id)
        return ''

    def render_PUT(self, request):
        logger.info("http put: %s" % request.path)
        user, blob_id = self._validate(request)
        d = self._handler.write_blob(user, blob_id, request)
        d.addCallback(lambda _: request.finish())
        d.addErrback(self._error, request)
        return NOT_DONE_YET

    def _error(self, e, request):
        logger.error('Error processing request: %s' % e.getErrorMessage())
        request.setResponseCode(500)
        request.finish()

    def _validate(self, request):
        for arg in request.postpath:
            if arg and not VALID_STRINGS.match(arg):
                raise Exception('Invalid blob resource argument: %s' % arg)
        return request.postpath


if __name__ == '__main__':
    # A dummy blob server
    # curl -X PUT --data-binary @/tmp/book.pdf localhost:9000/user/someid
    # curl -X GET -o /dev/null localhost:9000/user/somerandomstring
    from twisted.python import log
    import sys
    log.startLogging(sys.stdout)

    from twisted.web.server import Site
    from twisted.internet import reactor

    # parse command line arguments
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('--port', default=9000, type=int)
    parser.add_argument('--path', default='/tmp/blobs/user')
    args = parser.parse_args()

    root = BlobsResource("filesystem", args.path)
    # I picture somethink like
    # BlobsResource(backend="filesystem", backend_opts={'path': '/tmp/blobs'})

    factory = Site(root)
    reactor.listenTCP(args.port, factory)
    reactor.run()



# -*- coding: utf-8 -*-
# entrypoint.py
# Copyright (C) 2016 LEAP
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""
The entrypoint for Soledad server.

This is the entrypoint for the application that is loaded from the initscript
or the systemd script.
"""

from twisted.internet import reactor
from twisted.python import threadpool

from .auth import portalFactory
from .session import SoledadSession
from ._config import get_config
from ._wsgi import init_couch_state


# load configuration from file
conf = get_config()


class SoledadEntrypoint(SoledadSession):

    def __init__(self):
        pool = threadpool.ThreadPool(name='wsgi')
        reactor.callWhenRunning(pool.start)
        reactor.addSystemEventTrigger('after', 'shutdown', pool.stop)
        portal = portalFactory(pool)
        SoledadSession.__init__(self, portal)


# see the comments in application.py recarding why couch state has to be
# initialized when the reactor is running

reactor.callWhenRunning(init_couch_state, conf)
